python3 ./Test/astc_test_image.py --test-set all --block-size all --test-quality all --repeats 5 --encoder ref-master-avx2
python3 ./Test/astc_test_image.py --test-set all --block-size all --test-quality all --repeats 5 --encoder ref-master-sse4.2
python3 ./Test/astc_test_image.py --test-set all --block-size all --test-quality all --repeats 5 --encoder ref-master-sse2
